/* version1.0 (c) Copyright 2000 by Ian Wilson
   Written by Ian Wilson.
   Permission is granted to copy and use this program provided no fee is
   charged for it and provided that this copyright notice is not removed. */
#include "patlookup.h"
#include "util.h"
#include "nrutil.h"
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
/********************************************************/
/*  This file holds lookup tables for probabilities
    of the different triplets of parents and offspring  

  The mutational model is k alleles with  k being the 
  maximum number of alleles at a locus - this may not be
  perfect but if there are approximately the same 
  number of alleles at each locus will be OK            

  if NODATA is defined then the model prior should be 
   returned for the number of parents and their genotypes 
                                                        */
/********************************************************/
extern int max_alleles;
/**************************************************************/
double ***loghetlookup;
double ***loghomlookup;
double **logtryhetlookup;
double **logtryhomlookup;
int mutation_rates;
/***************************************************************/
double proboff(int i, int  k,double mu) 
/*    the offspring is 01 and transmitted alleles are i and k    */
{
    if (i + k == 1) {  /* ie both correct */ 
        return (1. - mu)*(1.-mu) 
			+ 2.*(1.-mu)*mu/(double)max_alleles 
			+2.*mu*mu/((double)max_alleles*(double)max_alleles);
    } else if (i + k == 4) {
        return 2.*mu*mu/((double)max_alleles*(double)max_alleles);
    } else if (i == k) {
        return 2.*mu*(1.- mu)/(double)max_alleles 
			+ 2.*mu*mu/((double)max_alleles*(double)max_alleles);
    } else {
        return mu * (1. - mu)/(double)max_alleles  + 
			2.*mu*mu/((double)max_alleles*(double)max_alleles);
    }
}
/**************************************************************/
double proboffhom(int i,int  k,double mu) 
/* the offspring is 00 and the parents i and k */
{
    if (i + k == 0) {
        return (1. - mu)*(1.-mu) 
					+ 2.*mu*(1-mu)/(double)max_alleles + 
					mu*mu/((double)max_alleles*(double)max_alleles);
    }
    else if (i + k == 2) {
        return mu*mu/((double)max_alleles*(double)max_alleles);
    }
    else {
        return mu * (1. - mu)/(double)max_alleles
			+ mu*mu/((double)max_alleles*(double)max_alleles);
    }
}
/**************************************************************/
void getloghetlookup(double **mat, double mu)
{
	int i,j,k,l;

	for (i=0;i<=2;i++) {				/* male[1]		*/
		for (j=0;j<=2;j++) {			/* male[2]		*/
			for (k=0;k<=2;k++) {		/* female[1]	*/
				for (l=0;l<=2;l++) {	/* female[2]	*/
					mat[3*i+j][3*k+l]=log((proboff(i,k,mu)+proboff(i,l,mu)
						+proboff(j,k,mu)+proboff(j,l,mu)))-2.*M_LN2;
				}
			}
		}
	}
}
/**************************************************************/
void getloghomlookup(double **mat, double mu)
{
	int i,j,k,l;
	
	for (i=0;i<=1;i++) {
		for (j=0;j<=1;j++) {
			for (k=0;k<=1;k++) {
				for (l=0;l<=1;l++) {
					mat[2*i+j][2*k+l]=log((proboffhom(i,k,mu)+proboffhom(i,l,mu)
						+proboffhom(j,k,mu)+proboffhom(j,l,mu)))-2.*M_LN2;
				}
			}
		}
	}
}
/**************************************************************/
double **allocate_lookup(int n)
{
	double **dtmp;
	int i;

    dtmp=(double **)MALLOC(n*sizeof(double*));
	if (!dtmp) {
		printf("error in setup_lookup");
		exit(0);
	}
	dtmp[0]=(double *)MALLOC(n*n*sizeof(double));
	if (!dtmp[0]) {
		printf("error in setup_lookup");
		exit(0);
	}
	for (i=1;i<n;i++) dtmp[i]=dtmp[i-1]+n;
	return dtmp;
}
/*********************************/
void setup_lookup(double mu, int nmu)
{
	int i;

	mutation_rates=nmu;
	printf("\n%d alleles used in k-allele model\n",max_alleles);
	
	loghetlookup=(double ***) MALLOC((nmu+1)*sizeof(double));
	if (!loghetlookup) exit(0);
	
	loghomlookup=(double ***) MALLOC((nmu+1)*sizeof(double));
	if (!loghomlookup) exit(0);
	
	for (i=1;i<=nmu;i++) {
		loghetlookup[i]=allocate_lookup(9);		
		loghomlookup[i]=allocate_lookup(4);		
		
		getloghetlookup(loghetlookup[i],mu);
		getloghomlookup(loghomlookup[i],mu);
	}
	logtryhetlookup=allocate_lookup(9);
	logtryhomlookup=allocate_lookup(4);
}
/*********************************/
void destroy_lookup(void)
{
	int i;

	FREE(logtryhomlookup[0]);FREE(logtryhomlookup);
	FREE(logtryhetlookup[0]);FREE(logtryhetlookup);
	for (i=1;i<=mutation_rates;i++) {
		FREE(loghomlookup[i][0]);FREE(loghomlookup[i]);
		FREE(loghetlookup[i][0]);FREE(loghetlookup[i]);	
	}
	FREE(loghomlookup);
	FREE(loghetlookup);
}
/**************************************************************/
void gettrylookup( double newmu)
{
	getloghetlookup(logtryhetlookup,newmu);
	getloghomlookup(logtryhomlookup,newmu);
}
/**************************************************************/
void getlookup( double newmu, int which)
{
	getloghetlookup(loghetlookup[which],newmu);
	getloghomlookup(loghomlookup[which],newmu);
}
/**************************************************************/
void swap_tables(int which)
{
	double **dtmp;
	dtmp=loghetlookup[which];
	loghetlookup[which]=logtryhetlookup;
	logtryhetlookup=dtmp;
	dtmp=loghomlookup[which];
	loghomlookup[which]=logtryhomlookup;
	logtryhomlookup=dtmp;
}
/**************************************************************/
double lookup_prob(int *gen, int *male, int *female, int which)
{
#ifdef NODATA
	return 0.0;
#else 
	int m,f;

	if (gen[1]<0||gen[2]<0) return 0.0;

	if (gen[1]==gen[2]) {
		if (male[1]==gen[1]) m=0;
		else  m=2;
		if (male[2]!=gen[1]) m+=1;
		if (female[1]==gen[1]) f=0;
		else  f=2;
		if (female[2]!=gen[1]) f+=1;
		return exp(loghomlookup[which][m][f]);
	} else {
		if (male[1]==gen[1]) m=0;
		else if (male[1]==gen[2]) m=3;
		else m=6;
		if (male[2]==gen[2]) m+=1;
		else if (male[2]!=gen[1]) m+=2;
		if (female[1]==gen[1]) f=0;
		else if (female[1]==gen[2]) f=3;
		else f=6;
		if (female[2]==gen[2]) f+=1;
		else if (female[2] !=gen[1]) f+=2;
		return exp(loghetlookup[which][m][f]);
	}
#endif
}
/**************************************************************/
double log_lookup_prob(int *gen, int *male, int *female, int which)
{
#ifdef NODATA
	return 0.0;
#else
	int m,f;

	if (gen[1]<=0||gen[2]<=0) return 0.0;	
	if (gen[1]==gen[2]) {
		if (male[1]==gen[1]) m=0;
		else  m=2;
		if (male[2]!=gen[1]) m+=1;
		if (female[1]==gen[1]) f=0;
		else  f=2;
		if (female[2]!=gen[1]) f+=1;
		return loghomlookup[which][m][f];
	} else {
		if (male[1]==gen[1]) m=0;
		else if (male[1]==gen[2]) m=3;
		else m=6;
		if (male[2]==gen[2]) m+=1;
		else if (male[2]!=gen[1]) m+=2;
		if (female[1]==gen[1]) f=0;
		else if (female[1]==gen[2]) f=3;
		else f=6;
		if (female[2]==gen[2]) f+=1;
		else if (female[2] !=gen[1]) f+=2;
		return loghetlookup[which][m][f];
	}
#endif
}
/**************************************************************/
double log_lookup_diff(int *gen, int *male, int *female, int which)
{
#ifdef NODATA
	return 0.0;
#else
	int m,f;
	
	if (gen[1]<=0||gen[2]<=0) return 0.0;

	if (gen[1]==gen[2]) {
		if (male[1]==gen[1]) m=0;
		else  m=2;
		if (male[2]!=gen[1]) m+=1;
		if (female[1]==gen[1]) f=0;
		else  f=2;
		if (female[2]!=gen[1]) f+=1;
		return logtryhomlookup[m][f]-loghomlookup[which][m][f];
	} else {
		if (male[1]==gen[1]) m=0;
		else if (male[1]==gen[2]) m=3;
		else m=6;
		if (male[2]==gen[2]) m+=1;
		else if (male[2]!=gen[1]) m+=2;
		if (female[1]==gen[1]) f=0;
		else if (female[1]==gen[2]) f=3;
		else f=6;
		if (female[2]==gen[2]) f+=1;
		else if (female[2] !=gen[1]) f+=2;
		return logtryhetlookup[m][f]-loghetlookup[which][m][f];
	}
#endif
}
    /* the next matrix is a lookup table for the number of mutations
	   for a homozygote offspring the hom is (0,0)
	   homozygote where the parents are ordered
	   (0,0), (0,1), (1,0), (1,1)                                 */

int hommutlookup[4][4]={
	{0,0,0,1},
	{0,0,0,1},
	{0,0,0,1},
	{1,1,1,2}
};
    /* the next matrix is a lookup table for the number of mutations
	   for a heterozygote offspring which has a (0,1) genotype
	   homozygote where the parents are ordered
	   (0,0),(0,1),(0,2),(1,0),(1,1),(1,2),(2,0),(2,1),(2,2)      */

int hetmutlookup[9][9]= 
{
	{1,0,1,0,0,0,1,0,1},
	{0,0,0,0,0,0,0,0,1},
	{1,0,1,0,0,0,1,0,1},
	{0,0,0,0,0,0,0,0,1},
	{0,0,0,0,1,1,0,1,1},
	{0,0,0,0,1,1,0,1,1},
	{1,0,1,0,0,0,1,0,1},
	{0,0,0,0,1,1,0,1,1},
	{1,1,1,1,1,1,1,1,2}
};
/**********************************************************/
int lookup_muts(int *gen, int *male, int *female)
{
	int m,f;

	if (gen[1]<=0||gen[2]<=0) return 0;
    /* a check here - it is not called often and should catch problems */
	if (male[1]<0||male[2]<0||female[1]<0||female[2]<0)
		mywarning("missing data in mothers or fathers");

	if (gen[1]==gen[2]) {
		if (male[1]==gen[1]) m=0;
		else  m=2;
		if (male[2]!=gen[1]) m+=1;
		if (female[1]==gen[1]) f=0;
		else  f=2;
		if (female[2]!=gen[1]) f+=1;
		return hommutlookup[m][f];
	} else {
		if (male[1]==gen[1]) m=0;
		else if (male[1]==gen[2]) m=3;
		else m=6;
		if (male[2]==gen[2]) m+=1;
		else if (male[2]!=gen[1]) m+=2;
		if (female[1]==gen[1]) f=0;
		else if (female[1]==gen[2]) f=3;
		else f=6;
		if (female[2]==gen[2]) f+=1;
		else if (female[2] !=gen[1]) f+=2;
		return hetmutlookup[m][f];
	}
}
/*******************************************************/

